# Copyright (c) 2008-2011, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

XA_PRIMARY = 1
XA_SECONDARY = 2
XA_ARC = 3
XA_ATOM = 4
XA_BITMAP = 5
XA_CARDINAL = 6
XA_COLORMAP = 7
XA_CURSOR = 8
XA_CUT_BUFFER0 = 9
XA_CUT_BUFFER1 = 10
XA_CUT_BUFFER2 = 11
XA_CUT_BUFFER3 = 12
XA_CUT_BUFFER4 = 13
XA_CUT_BUFFER5 = 14
XA_CUT_BUFFER6 = 15
XA_CUT_BUFFER7 = 16
XA_DRAWABLE = 17
XA_FONT = 18
XA_INTEGER = 19
XA_PIXMAP = 20
XA_POINT = 21
XA_RECTANGLE = 22
XA_RESOURCE_MANAGER = 23
XA_RGB_COLOR_MAP = 24
XA_RGB_BEST_MAP = 25
XA_RGB_BLUE_MAP = 26
XA_RGB_DEFAULT_MAP = 27
XA_RGB_GRAY_MAP = 28
XA_RGB_GREEN_MAP = 29
XA_RGB_RED_MAP = 30
XA_STRING = 31
XA_VISUALID = 32
XA_WINDOW = 33
XA_WM_COMMAND = 34
XA_WM_HINTS = 35
XA_WM_CLIENT_MACHINE = 36
XA_WM_ICON_NAME = 37
XA_WM_ICON_SIZE = 38
XA_WM_NAME = 39
XA_WM_NORMAL_HINTS = 40
XA_WM_SIZE_HINTS = 41
XA_WM_ZOOM_HINTS = 42
XA_MIN_SPACE = 43
XA_NORM_SPACE = 44
XA_MAX_SPACE = 45
XA_END_SPACE = 46
XA_SUPERSCRIPT_X = 47
XA_SUPERSCRIPT_Y = 48
XA_SUBSCRIPT_X = 49
XA_SUBSCRIPT_Y = 50
XA_UNDERLINE_POSITION = 51
XA_UNDERLINE_THICKNESS = 52
XA_STRIKEOUT_ASCENT = 53
XA_STRIKEOUT_DESCENT = 54
XA_ITALIC_ANGLE = 55
XA_X_HEIGHT = 56
XA_QUAD_WIDTH = 57
XA_WEIGHT = 58
XA_POINT_SIZE = 59
XA_RESOLUTION = 60
XA_COPYRIGHT = 61
XA_NOTICE = 62
XA_FONT_NAME = 63
XA_FAMILY_NAME = 64
XA_FULL_NAME = 65
XA_CAP_HEIGHT = 66
XA_WM_CLASS = 67
XA_WM_TRANSIENT_FOR = 68
XA_LAST_PREDEFINED = 68

KeyPress = 2
KeyRelease = 3
ButtonPress = 4
ButtonRelease = 5
MotionNotify = 6
EnterNotify = 7
LeaveNotify = 8
FocusIn = 9
FocusOut = 10
KeymapNotify = 11
Expose = 12
GraphicsExpose = 13
NoExpose = 14
VisibilityNotify = 15
CreateNotify = 16
DestroyNotify = 17
UnmapNotify = 18
MapNotify = 19
MapRequest = 20
ReparentNotify = 21
ConfigureNotify = 22
ConfigureRequest = 23
GravityNotify = 24
ResizeRequest = 25
CirculateNotify = 26
CirculateRequest = 27
PropertyNotify = 28
SelectionClear = 29
SelectionRequest = 30
SelectionNotify = 31
ColormapNotify = 32
ClientMessage = 33
MappingNotify = 34
LASTEvent = 35



# Copyright (c) 2008-2011, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
    This module contains some helper functions and methods
    dealing with the `ewmh standard`_.
    They can also mixed into their corresponding classes.

    .. _ewmh standard: http://standards.freedesktop.org/wm-spec/wm-spec-latest.html
"""

from ooxcb.list import List

def ewmh_get_client_list(screen):
    """
        return a list of :class:`ooxcb.protocol.xproto.Window` instances, or
        None if the _NET_CLIENT_LIST property is not set.
    """
    reply = screen.root.get_property('_NET_CLIENT_LIST', 'WINDOW').reply()
    if not reply.exists:
        return None
    else:
        return reply.value.to_windows()

def ewmh_get_client_list_stacking(screen):
    """
        return a list of :class:`ooxcb.protocol.xproto.Window` instances, or
        None if the _NET_CLIENT_LIST_STACKING property is not set.
    """
    reply = screen.root.get_property('_NET_CLIENT_LIST_STACKING', 'WINDOW').reply()
    if not reply.exists:
        return None
    else:
        return reply.value.to_windows()

def ewmh_get_number_of_desktops(screen):
    """
        return the number of desktops, or None if the property
        _NET_NUMBER_OF_DESKTOPS is not set.
    """
    reply = screen.root.get_property('_NET_NUMBER_OF_DESKTOPS', 'CARDINAL').reply()
    if not reply.exists:
        return None
    else:
        return reply.value[0]

def ewmh_get_current_desktop(screen):
    """
        return the current desktop index (starting at 0),
        or None if the _NET_CURRENT_DESKTOP property is not set.
    """
    reply = screen.root.get_property('_NET_CURRENT_DESKTOP', 'CARDINAL').reply()
    if not reply.exists:
        return None
    else:
        return reply.value[0]

def ewmh_get_desktop(window):
    """
        return the desktop of the window or None
        if the _NET_WM_DESKTOP property is not set.
        0xffffffff is returned if the window wants to be
        visible on all desktops (according to the ewmh spec)
    """
    reply = window.get_property('_NET_WM_DESKTOP', 'CARDINAL').reply()
    if not reply.exists:
        return None
    else:
        return reply.value[0]

# TODO emwh_get_visible_window_name/get_window_name
def ewmh_get_window_name(window):
    """
        returns the window title you should use.
        Since it respects the icccm and the ewmh standard,
        it will use:

         * _NET_WM_VISIBLE_NAME if available. if not,
         * _NET_WM_NAME if available. if not,
         * WM_NAME

        If `WM_NAME` is not available, it will return an empty string.

        .. note:: WM_NAME's encoding is latin-1, _NET_WM_NAME and
                _NET_WM_VISIBLE_NAME are utf-8-encoded.
    """
    # try _NET_WM_VISIBLE_NAME
    encoding = 'utf-8'
    reply = window.get_property(
            '_NET_WM_VISIBLE_NAME', 'UTF8_STRING').reply()
    if not reply.exists:
        # try _NET_WM_NAME
        reply = window.get_property(
                '_NET_WM_NAME', 'UTF8_STRING').reply()
        if not reply.exists:
            # use WM_NAME ( has ... to ... exist! )
            reply = window.get_property(
                    'WM_NAME', 'STRING').reply()
            encoding = 'latin-1'
            if not reply.exists:
                return ''
    return reply.value.to_string().decode(encoding)


def ewmh_set_window_name(window, name):
    if type(name) is not unicode:
        name = unicode(name, 'utf8', 'replace')
    window.change_property('WM_NAME', 'STRING', 8, List.from_string(name.encode('latin-1', 'replace')))
    window.change_property('_NET_WM_NAME', 'UTF8_STRING', 8, List.from_string(name.encode('utf8', 'replace')))


def mixin():
    """
        mix em all
    """
    from ooxcb.util import mixin_functions
    from ooxcb.protocol.xproto import Screen, Window
    
    mixin_functions((
        ewmh_get_client_list,
        ewmh_get_client_list_stacking,
        ewmh_get_number_of_desktops,
        ewmh_get_current_desktop,
        ), Screen)
    mixin_functions((
        ewmh_get_desktop,
        ewmh_get_window_name,
        ewmh_set_window_name,
        ), Window)



# auto generated. yay.
import ooxcb
from ooxcb.resource import get_internal
from ooxcb.types import SIZES, make_array, build_list
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
from struct import pack, unpack, calcsize
from ooxcb.protocol.xproto import Window
from ooxcb.util import Mixin

def unpack_from_stream(fmt, stream, offset=0):
    if offset:
        stream.seek(offset, 1)
    s = stream.read(calcsize(fmt))
    return unpack(fmt, s)


MAJOR_VERSION = 2
MINOR_VERSION = 1
key = ooxcb.ExtensionKey("XTEST")

class Cursor(object):
    _None = 0
    Current = 1

class GetVersionReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.major_version = None
        self.minor_version = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xBxxxxxxH", stream)
        self.major_version = _unpacked[0]
        self.minor_version = _unpacked[1]

    def build(self, stream):
        count = 0
        stream.write(pack("=xBxxxxxxH", self.major_version, self.minor_version))

class WindowMixin(Mixin):
    target_class = Window
    def compare_cursor(self, cursor):
        window = get_internal(self)
        cursor = get_internal(cursor)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", window, cursor))
        return self.conn.xtest.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, False, True), \
            CompareCursorCookie(),
            CompareCursorReply)

    def compare_cursor_unchecked(self, cursor):
        window = get_internal(self)
        cursor = get_internal(cursor)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxII", window, cursor))
        return self.conn.xtest.send_request(ooxcb.Request(self.conn, buf.getvalue(), 1, False, False), \
            CompareCursorCookie(),
            CompareCursorReply)

class xtestExtension(ooxcb.Extension):
    header = "xtest"
    def get_version(self, major_version, minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBxH", major_version, minor_version))
        return self.conn.xtest.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, True), \
            GetVersionCookie(),
            GetVersionReply)

    def get_version_unchecked(self, major_version, minor_version):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBxH", major_version, minor_version))
        return self.conn.xtest.send_request(ooxcb.Request(self.conn, buf.getvalue(), 0, False, False), \
            GetVersionCookie(),
            GetVersionReply)

    def fake_input_checked(self, type, detail=0, time=0, window=None, rootX=0, rootY=0, deviceid=0):
        window = get_internal(window)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBxxIIxxxxxxxxHHxxxxxxxB", type, detail, time, window, rootX, rootY, deviceid))
        return self.conn.xtest.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, True), \
            ooxcb.VoidCookie())

    def fake_input(self, type, detail=0, time=0, window=None, rootX=0, rootY=0, deviceid=0):
        window = get_internal(window)
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBBxxIIxxxxxxxxHHxxxxxxxB", type, detail, time, window, rootX, rootY, deviceid))
        return self.conn.xtest.send_request(ooxcb.Request(self.conn, buf.getvalue(), 2, True, False), \
            ooxcb.VoidCookie())

    def grab_control_checked(self, impervious):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBxxx", impervious))
        return self.conn.xtest.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, True), \
            ooxcb.VoidCookie())

    def grab_control(self, impervious):
        buf = StringIO.StringIO()
        buf.write(pack("=xxxxBxxx", impervious))
        return self.conn.xtest.send_request(ooxcb.Request(self.conn, buf.getvalue(), 3, True, False), \
            ooxcb.VoidCookie())

class CompareCursorReply(ooxcb.Reply):
    def __init__(self, conn):
        ooxcb.Reply.__init__(self, conn)
        self.same = None

    def read(self, stream):
        self._address = stream.address
        _unpacked = unpack_from_stream("=xBxxxxxx", stream)
        self.same = _unpacked[0]

    def build(self, stream):
        count = 0
        stream.write(pack("=xBxxxxxx", self.same))

class CompareCursorCookie(ooxcb.Cookie):
    pass

class GetVersionCookie(ooxcb.Cookie):
    pass

_events = {
}

_errors = {
}

for ev in _events.itervalues():
    if isinstance(ev.event_target_class, str):
        ev.event_target_class = globals()[ev.event_target_class]

ooxcb._add_ext(key, xtestExtension, _events, _errors)
def mixin():
    WindowMixin.mixin()





# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ctypes

class MemoryInputStream(object):
    """
        `MemoryInputStream` is an object implementing the buffer
        interface which reads its data from a certain
        memory buffer. It is interpreted as an infinite
        array of 8-bit chars.
        This stream is used for ooxcb protocol parsing.
    """
    def __init__(self, address):
        """
            :Parameters:
                `address` : int
                    The address of the first character.
        """
        self._root = address
        self._ptr = address
        self._opened = True

    def close(self):
        self._opened = False

    @property
    def closed(self):
        return not self._opened

    def fileno(self):
        """ We are not using a file descriptor """
        raise IOError()

    def flush(self):
        pass

    def isatty(self):
        return False

    def readable(self):
        return True

    def readline(self, limit):
        raise NotImplementedError()

    def readlines(self, hint):
        raise NotImplementedError()

    def seek(self, offset, whence=0):
        """
            Currently you can not use whence=2.
            The stream has no end.
        """
        if whence == 0:
            # from the start position of the stream
            self._ptr = self._root + offset
            return offset
        elif whence == 1:
            # from the current position of the stream
            self._ptr += offset
            return self._ptr - self._root
        elif whence == 2:
            # from the end of the stream ... impossible, sorry.
            raise NotImplementedError('The memory stream has no end!')
        else:
            raise NotImplementedError('Unknown whence: %d' % whence)

    @property
    def address(self):
        return self._root

    def seekable(self):
        return True

    def tell(self):
        return self._ptr - self._root

    def truncate(self):
        """
            That's a no-op here. I can't think of anyone needing that
            for a memory buffer, really.
            However, it will raise a NotImplementedError. But we can
            remove that.
        """
        raise NotImplementedError() # hmmm ...

    def writable(self):
        return False

    def write(self, b):
        raise IOError()

    def read(self, b=None):
        if b is None:
            raise NotImplementedError('The memory stream has no end!')
        chunk = ctypes.string_at(self._ptr, b)
        self._ptr += b
        return chunk

    def readall(self):
        raise NotImplementedError('The memory stream has no end!')

    def readinto(self, b):
        length = len(b)
        b.extend(self.read(length))
        return length

if __name__ == '__main__':
    # TODO: build a nice test for that.
    buf = ctypes.create_string_buffer('abcdefghijklmnopqrstuvwxyz')
    stream = MemoryInputStream(ctypes.addressof(buf))
    print stream.read(2)
    print stream.read(4)
    print stream.seek(0)
    print stream.read(26)
    print repr(stream.read(10))




# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .exception import ProtocolDataError
from .eventsys import EventDispatcher

class Resource(EventDispatcher):
    """
        A resource is nearly every object in the X world having
        an X ID: Windows, Graphics Contexts, Fonts ...
        Except atoms. Atoms are not part of the XID space, and
        they have a separate cache.

        Each resource has an *xid* attribute. It can also be accessed
        by :meth:`get_internal`.

        Resources are also :class:`ooxcb.eventsys.EventDispatcher`
        subclasses.
    """
    def __init__(self, conn, xid):
        self.conn = conn
        assert isinstance(xid, (int, long)), \
                "You have specified an invalid X ID: %s" % xid
        self.xid = xid

    def __repr__(self):
        return '<%s XID=%d (0x%x)>' % (
                self.__class__.__name__, self.xid, id(self)
                )

    def get_internal(self):
        """
            The internal representation of a resource is its X id.
            Return it.
        """
        return self.xid

    # No __hash__ / __cmp__ should be needed here. Resources are cached,
    # identity comparison should be sufficient. However, TODO?

class _XNone(object):
    """
        A class for *XNone*. That's something with 0 as an internal
        representation and often used :)
    """
    def __repr__(self):
        return '<XNone>'

    def get_internal(self):
        return 0

XNone = _XNone()

def get_internal(obj):
    """
        get the stream-ready internal representation of *obj*.
        There are multiple possible cases:

            * If *obj* has a `get_internal` method, the return value of
              that will be returned.
            * If *obj* is an int, it will be returned immediately.
            * If *obj* is None, 0 will be returned.
            * If *obj* is something else, a
              :class:`ProtocolDataError <ooxcb.ProtocolDataError>` will be
              raised.
    """
    if hasattr(obj, 'get_internal'):
        return obj.get_internal()
    elif isinstance(obj, int):
        return obj
    elif obj is None:
        return 0
    else:
        raise ProtocolDataError("The internal value of %r couldn't be retrieved." % obj)




# Copyright (c) 2008-2010, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ctypes

from . import libc
from .libxcb import xcb_generic_reply_t
from .util import cached_property
from .memstream import MemoryInputStream

class Protobj(object):
    """
        Base class for all protocol objects.

        .. data:: create_lazy

            None as default, but it can be a classmethod used
            to create a Protobj instance lazily for convenience.

            :see: :func:`ooxcb.builder.build_list`

        .. data:: pythonize_lazy

            None as default, but if it can be a method used to
            pythonize a protocol object lazily if it's inside
            a list.

            :see: :class:`ooxcb.list.List`

    """
    create_lazy = None # TODO: does it have to be defined here?
    pythonize_lazy = None

    def __init__(self, conn):
        self.conn = conn

    def read(self, stream):
        """
            Placeholder for subclasses.
            :param stream: an object providing a buffer interface. Read
                           myself from that stream!
        """
        raise NotImplementedError()

    def read_from_address(self, address):
        """
            parse the memory at *address*
        """
        return self.read(MemoryInputStream(address))

    @classmethod
    def create_from_address(cls, conn, address):
        """
            :type conn: :class:`ooxcb.conn.Connection`
            :param address: an int pointing to the data in the
                            memory
        """
        self = cls(conn)
        self.read_from_address(address)
        return self

    @classmethod
    def create_from_stream(cls, conn, stream):
        """
            :type conn: :class:`ooxcb.conn.Connection`
            :param stream: a stream-like object to read from.
        """
        self = cls(conn)
        self.read(stream)
        return self

class Struct(Protobj):
    """
        also just a subclass of :class:`protobj.Protobj`. Nothing special here.
    """
    pass

class Union(Protobj):
    """
        the baseclass for Unions. Nothing special.
    """
    pass

class Request(Protobj):
    """
        mostly just a value holder for :meth:`ooxcb.ext.Extension.send_request`
    """
    def __init__(self, conn, buffer, opcode, void, checked):
        # TODO: check size??
        self.opcode = opcode
        self.is_void = void
        self.is_checked = checked

        Protobj.__init__(self, conn)

        self.buffer = buffer
        self.size = len(buffer)

class Response(Protobj):
    """
        a response wrapper. Had a *sequence* getter in the past,
        but it isn't used. If it should be added again, file a bug.
    """
    pass

class Error(Response):
    """
        An error is a special kind of response that is used when an
        error occured. It will be "packed into" a Python exception
        in :meth:`Error.set`.
    """
    @classmethod
    def set(cls, conn, err):
        """
            raise the right exception for the
            :class:`ooxcb.libxcb.xcb_generic_error_t` struct *err*
            if it is an error.

            :param err: a ctypes pointer to a
                        :class:`ooxcb.libxcb.xcb_generic_error_t`.
        """
        if err:
            e = err.contents
            # get the opcode and the corresponding error type / exception class
            opcode = e.error_code
            type, exception = conn.errors[opcode]
            address = ctypes.addressof(e)

            inst = type.create_from_address(conn, address)
            libc.free(address)

            raise exception(conn, inst)

class Event(Response):
    """
        An event is a special kind of response.
        And in ooxcb, it comes with a builtin ability to be dispatched.

        For that purpose, every Event subclass has an `event_name` member.
        That's the name of the event, e.g. 'on_configure_notify'. It is
        'on_event' as fallback here, so if you stumble upon 'on_event'
        somewhere, there is something wrong.

        So, if you want to dispatch an event, just call :meth:`dispatch`.
        A typical line in the main loop would then be:

        ::

            conn.wait_for_event().dispatch()

        which will automatically dispatch every received event.

        Every event will be dispatched to an appropriate target, e.g.
        a :class:`ooxcb.xproto.ButtonPressEvent` will be dispatched
        to the window the event happened in. The targets are defined
        in the interface files. The fallback target is the connection.

        If you want to know which event types a class receives, you can
        access the *event_types* attribute that is a list of event names.

        If you want to know the event target of an event, you can access
        its *event_target* attribute.

        Every event is sent with the event itself as the first and only
        argument. So all ooxcb event handlers have the following signature:

        ::

            def on_blabla(evt)

    """
    event_name = 'on_event'

    def __init__(self, conn):
        Response.__init__(self, conn)
        self.event_target = conn

    @classmethod
    def create(cls, conn, event):
        opcode = event.contents.response_type & ~0x80 # strip 'send event' bit

        type = cls

        if (opcode in conn.events and conn.events[opcode]):
            type = conn.events[opcode]

        address = ctypes.addressof(event.contents)
        evt = type.create_from_address(conn, address)
        libc.free(address)
        return evt

    def dispatch(self):
        """
            dispatch *self* to my event target
        """
        return self.event_target.dispatch_event(self.event_name, self)

class Reply(Response):
    """
        a reply is a :class:`response.Response` subclass.
    """
    @cached_property
    def _struct(self):
        return xcb_generic_reply_t.from_address(self._address)

    @property
    def length(self):
        """
            The reply length.
        """
        return self._struct.length




# Copyright (c) 2008-2011, samurai-x.org
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the samurai-x.org nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SAMURAI-X.ORG ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL SAMURAI-X.ORG  BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class AtomDict(dict):
    """
        A dictionary which is able to lazily load an atom
        and that contains all predefined atoms from :mod:`ooxcb.constant`.

        ::

            dic = AtomDict(my_connection)
            print dic['WM_CLASS'] # Yay, it is lazily loaded!
            aid = dic['WM_CLASS'].get_internal()
            # ... and vice-versa:
            assert dic.get_by_id(aid) == dic['WM_CLASS']

        You should not modify the dictionary manually.

    """
    def __init__(self, conn, *boo, **far):
        dict.__init__(self, *boo, **far)
        self.conn = conn
        self._by_id = {}
        self._add_predefined()
        self.do_name_lookup = True

    def _add_predefined(self):
        """
            add the predefined atoms from :module:`ooxcb.constant`.
        """
        # TODO: uglyuglyuglyuglyuglyUGLY
        from . import constant
        from ooxcb.protocol.xproto import Atom # TODO:
        for name in dir(constant):
            if name.startswith('XA_'):
                key = name[len('XA_'):]
                value = getattr(constant, name)
                atom = Atom(self.conn, value)
                self.add_atom(key, atom)

    def add_atom(self, name, atom):
        """
            add the atom *atom* with the name *name* to the cache.
        """
        self[name] = atom
        self._by_id[atom.get_internal()] = atom

    def __missing__(self, key):
        self[key] = value = self.conn.core.intern_atom(key, False).reply().atom
        self._by_id[value.get_internal()] = value
        return value

    def get_by_id(self, aid):
        """
            return an atom instance for the ID *aid*.

            That is not part of the connection's XID cache because atom ids
            do not seem to be part of the XID space; they have their own
            space.

            For aid == 0, it will return None, but this may
            change in the future.

            :todo: currently, this depends on :mod:`ooxcb.protocol.xproto`. That's
                   not really consistent.
        """
        from ooxcb.protocol.xproto import Atom # TODO: uuuuuuuugly
        if aid == 0:
            return None # TODO: That's basically AnyProperty ... better solution?
        try:
            return self._by_id[aid]
        except KeyError:
            atom = Atom(self.conn, aid)
            if self.do_name_lookup:
                self._by_id[aid] = atom
                name = atom.get_name().reply().name
                self[name] = atom
            return atom

